
(* splitting functions using [Str] *)

let int_of_string s =
  try
    int_of_string s
  with
      Failure _ -> failwith ("int_of_string " ^ s)
;;
	

let spaces = Str.regexp "[ ]+"
let split s = Str.split spaces s
let split_ints s = List.map int_of_string (split s)
let sharp = Str.regexp "#"
let split_sharp s = Str.split sharp s

type direction = North | South | East | West

let char_of_direction d =
  match d with
  | North -> 'N'
  | East -> 'E'
  | West -> 'W'
  | South -> 'S'

(* initialization *)

let _ = Connect.send_line "Player"

(* the board: an array of [(width+2) * (height+2)] of cells;
   walls are set on the four edges *)
let width, height = 
  let s = Connect.read_line () in
  match split_ints s with [w; h] -> w, h | _ -> assert false

type cell = Plain | Water | Wall | Home | Robot

let board = Array.create_matrix (width + 2) (height + 2) Plain

let bases_ = ref []

(* getting the board *)
let _ =
  for row = 1 to height do
    board.(0).(row) <- Wall; board.(width + 1).(row) <- Wall;
    let s = Connect.read_line () in
    (* prerr_endline s; *)
    for col = 1 to width do
      match s.[col - 1] with
	| '.' -> ()
	| '~' -> board.(col).(row) <- Water
	| '#' -> board.(col).(row) <- Wall
	| '@' -> board.(col).(row) <- Home;
	         bases_ := (col, row) :: !bases_
	| _ -> assert false
    done
  done;
  for col = 0 to width + 1 do
    board.(col).(0) <- Wall; board.(col).(height + 1) <- Wall
  done

(* the list of bases' coordinates *)
let bases = !bases_

(* packages' info *)
type package_info = { dest_x : int; dest_y : int; weight : int }
let packages : (int, package_info) Hashtbl.t = Hashtbl.create 97

let package_weight p = 
  try (Hashtbl.find packages p).weight with Not_found -> 0

(* packages available on the board *)
let available = Array.create_matrix (width + 2) (height + 2) Ptset.empty
let available_packages = Hashtbl.create 257

(* the player id, its capacity and its initial money *)
let self_id, max_capacity, init_money =
  let s = Connect.read_line () in
  match split_ints s with [i; c; m] -> i, c, m | _ -> assert false

let money = ref init_money

let capacity = ref max_capacity

(* all the players are stored in hash table [players] *)
type player = { 
  mutable x : int;
  mutable y : int; 
  mutable load : Ptset.t (* set of carried packages *)
}

let players = Hashtbl.create 97

(* nothing changed regarding the current path *)
let next_position = ref (0,0)

let nothing_changed = ref true

(* interpreting one player action *)
let interp_player_reply player = 
  let p = 
    try 
      Hashtbl.find players player 
    with Not_found ->
      let p = { x = 0; y = 0; load = Ptset.empty } in
      Hashtbl.add players player p; p
  in
  let rec interp = function
    | "X" :: x :: r -> p.x <- int_of_string x; interp r
    | "Y" :: y :: r -> p.y <- int_of_string y; interp r
    | "N" :: r -> p.y <- succ p.y; interp r
    | "S" :: r -> p.y <- pred p.y; interp r
    | "E" :: r -> p.x <- succ p.x; interp r
    | "W" :: r -> p.x <- pred p.x; interp r
    | "P" :: pi :: r -> 
	nothing_changed := false;
	let pi = int_of_string pi in
	p.load <- Ptset.add pi p.load; 
	if player == self_id then capacity := !capacity - package_weight pi;
	available.(p.x).(p.y) <- Ptset.remove pi available.(p.x).(p.y);
	Hashtbl.remove available_packages pi;
	interp r
    | "D" :: pi :: r -> 
	begin
	  try
	let pi = int_of_string pi in
	p.load <- Ptset.remove pi p.load; 
	if player == self_id then capacity := !capacity + package_weight pi;
	(* Claude : BUG !!! 
	   devient available que si ce n'est pas sa destination finale ! *) 
	(* correction : *)
	begin
	  try
	    let pinfo = Hashtbl.find packages pi in
	    if p.x <> pinfo.dest_x || p.y <> pinfo.dest_y then
	      begin
		nothing_changed := false;
		available.(p.x).(p.y) <- Ptset.add pi available.(p.x).(p.y);
		Hashtbl.add available_packages pi (p.x,p.y);
	      end;	
	  with
	    Not_found ->
	      nothing_changed := false;
	      (* un robot a drope un paquet inconnu ! que fait-on ? *)
	      available.(p.x).(p.y) <- Ptset.add pi available.(p.x).(p.y);
	      Hashtbl.add available_packages pi (p.x,p.y);
	end;	
	  with Failure _ -> assert false;
	end;
	interp r
    | [] -> ()
    | _ -> assert false
  in
  interp

(* robots's places *)
let robots = Array.create_matrix (width + 2) (height + 2) 0

(* interpreting the server's reply *)
let interp_reply s =
  nothing_changed := true;
  Hashtbl.iter (fun _ p -> robots.(p.x).(p.y) <- 0) players;
  let allp = Hashtbl.fold (fun i _ l -> Ptset.add i l) players Ptset.empty in
  let seen = ref Ptset.empty in
  let update_one_player p = match split p with
    | i :: r -> 
	begin
	  try
	    let i = int_of_string i in
	    seen := Ptset.add i !seen;
	    interp_player_reply i r
	  with
	      Failure _ ->
		if i = "Robot" then
		  prerr_endline "It seems I died"
		else
		  prerr_endline ("server protocol not understood: " ^ p);
		raise Exit
	end	    
    | _ -> 
	assert false
  in
  List.iter update_one_player (split_sharp s);
  Ptset.iter 
    (fun i -> if not (Ptset.mem i !seen) then Hashtbl.remove players i) allp;
  Hashtbl.iter (fun i p -> robots.(p.x).(p.y) <- i) players

(* interpreting the initial situation (allocates the players) *)
let _ =
  let s = Connect.read_line () in
  interp_reply s

(* state of the robot *)

let old2_x = ref 0;;
let old2_y = ref 0;;
let old1_x = ref 0;;
let old1_y = ref 0;;

let self = Hashtbl.find players self_id
